/*=========================================================================

Program:   Insight Segmentation & Registration Toolkit
Module:    $RCSfile: itkekFMMEvaluateMultiquadrics3D.txx,v $
Language:  C++
Date:      $Date: 2004/08/06 12:38:33 $
Version:   $Revision: 1.38 $

Copyright (c) Insight Software Consortium. All rights reserved.
See ITKCopyright.txt or http://www.itk.org/HTML/Copyright.htm for details.

This software is distributed WITHOUT ANY WARRANTY; without even 
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef _evaluator_cxx
#define _evaluator_cxx

#include <limits>
#include <algorithm>
//#include <fenv.h>
#include <cassert>

#include "evaluator.h"
#include "util.h"

namespace fmi
{



   evaluator::
   evaluator()
   {
      m_Tau = 0.0;
      m_Tau2 = m_Tau*m_Tau;
      m_Depth = 2;
      m_K = 1;
      m_P = 5;
      m_PplusK = m_P + m_K;

      m_Error = 1.0e-6;

      m_ExtraPolynomialDegree = -1;

      // Purposely make the bounding box invalid
      // so we know to update it.
      m_BoundingBox[0]=1.0;
      m_BoundingBox[1]=0.0;
      m_BoundingBox[2]=1.0;
      m_BoundingBox[3]=0.0;
      m_BoundingBox[4]=1.0;
      m_BoundingBox[5]=0.0;

      m_NumberOfPoints = 0;

      m_Modified = true;
      m_ModifiedCoefficients = true;


      //m_PointsAreSorted = false;

      m_WellSeparatedFactor2 = 4.0;
      m_MinDegree = 0;
      m_DecreaseFactor = 1.85;
      m_DecreaseFactor2 = -0.85;

      // Set the cuttoff points. This array tells you for each degree how many
      // points should the box contain such that it's worth it to evaluate the
      // far field. This is purely empirical and needs to be tweaked.
      m_NumberCutoffPoints.resize(25);

      m_NumberCutoffPoints[0] = 2;
      m_NumberCutoffPoints[1] = 3;
      m_NumberCutoffPoints[2] = 4;
      m_NumberCutoffPoints[3] = 5;
      m_NumberCutoffPoints[4] = 6;
      m_NumberCutoffPoints[5] = 8;
      m_NumberCutoffPoints[6] = 12;
      m_NumberCutoffPoints[7] = 15;
      m_NumberCutoffPoints[8] = 21;
      m_NumberCutoffPoints[9] = 27;
      m_NumberCutoffPoints[10] = 35;
      m_NumberCutoffPoints[11] = 43;
      m_NumberCutoffPoints[12] = 53;
      m_NumberCutoffPoints[13] = 63;
      m_NumberCutoffPoints[14] = 73;
      m_NumberCutoffPoints[15] = 83;
      m_NumberCutoffPoints[16] = 93;
      m_NumberCutoffPoints[17] = 103;
/*
      for(unsigned int i=0;i<m_NumberCutoffPoints.size();i++)
      {
	 m_NumberCutoffPoints[i] = Polynomial::Offsets[i+1]/7;
//	 PRINTND(m_NumberCutoffPoints[i]);
      }
*/
//   The offsets are:
//
//      0, 1, 4, 10, 20, 35, 56, 84, 120, 165, 220, 286, 364, 455, 560, 680, 816,
//	 969, 1140, 1330, 1540, 1771, 2024
   }
   


   void
   evaluator::
   SetPointSet(const ScalarType* const landmarks, const int n)
   {

      if(!landmarks)
      {
	 std::cout << "Error. landmarks is null" << std::endl;
	 abort();
      }

      m_Points.resize(n);

      PointType p;

      for(int i=0;i<n;i++)
      {
	 p[0] = landmarks[3*i];
	 p[1] = landmarks[3*i+1];
	 p[2] = landmarks[3*i+2];
	 m_Points[i] = p;
      }      

      m_NumberOfPoints = n;

      // If the points are modified the bounding box may no longer be correct.
      // Therefore invalidate it in order to force a recomputation of the
      // bounding box again later on (in the funcion Update).
      m_BoundingBox[0]=1.0;
      m_BoundingBox[1]=0.0;
      m_BoundingBox[2]=1.0;
      m_BoundingBox[3]=0.0;
      m_BoundingBox[4]=1.0;
      m_BoundingBox[5]=0.0;

      //m_PointsAreSorted = pointsAreSorted;

      m_Modified = true;
      
   }



   void
   evaluator::
   SetCoefficients(const ScalarType* const coefficients, const int n)
   {
      if(!coefficients)
      {
	 std::cout << "Error. coefficients is null" << std::endl;
	 abort();
      }

      m_Coefficients.resize(n);
      for(int i=0;i<n;i++)
	 m_Coefficients[i] = coefficients[i];

      m_ModifiedCoefficients = true;
   }

   
   // Default is 2
   void 
   evaluator::
   SetDepth(int depth) 
   { 
      if(m_Depth != depth)
      {
	 m_Depth = depth; 
	 m_Modified = true; 
      }
   }

   // Note: this function assumes points are distributed uniformaly. If not
   // call SetDepth.
   void 
   evaluator::
   SetDepthToOptimal()
   { 

      if(m_NumberOfPoints <= 0)
      {
	 std::cout << "Error: Number of points is zero. Make sure to call "
	    "SetPointSet() before calling this function." << std::endl;
	 abort();
      }	 



      const int originalDepth = m_Depth;

/*
      ScalarType averagePointsPerBox = 15.0;

      m_Depth = static_cast<int>( 
	 floor( 
	    log(static_cast<ScalarType>(m_NumberOfPoints)/averagePointsPerBox) /
	    log(8.0) 
	    ) 
	 );

      if(m_Depth < 2)
	m_Depth = 2;

      std::cout << "Optimal depth: " << m_Depth << std::endl;
*/

      if(m_NumberOfPoints < 10000)
	 m_Depth = 2;
      else if(m_NumberOfPoints < 100000)
	 m_Depth = 3;
      else if(m_NumberOfPoints < 1000000)
	 m_Depth = 4;
      else if(m_NumberOfPoints < 10000000)
	 m_Depth = 5;
      else
      {
	 std::cout << "Error. Too many points." << std::endl;
	 abort();
      }
//      else if(m_NumberOfPoints < 10000)
//	 m_Depth = 2;


      // Only set the modification flag if the depth changed
      if(originalDepth != m_Depth)
	 m_Modified = true; 
   }

   // Default is 5
   void 
   evaluator::
   SetP(int p) 
   { 
      if(m_P != p)
      {
	 m_P = p; 
	 m_PplusK = m_P + m_K;
	 m_Modified = true; 
      }
   }
   
   // Default is 1
   void 
   evaluator::
   SetK(int k) 
   { 
      if(m_K != k)
      {
	 m_K = k; 
	 m_PplusK = m_P + m_K;
	 m_Modified = true; 
      }
   }

   // Default is 0.0
   void 
   evaluator::
   SetTau(ScalarType t) 
   { 
      if(m_Tau != t)
      {
	 m_Tau = t; 
	 m_Tau2 = m_Tau * m_Tau;
	 m_Modified = true; 
      }
   }

   void 
   evaluator::
   SetError(ScalarType e)
   {

      if(m_Error != e)
      {
	 m_Error = e;
	 m_Modified = true; 
	 
      }
      
   }


   void 
   evaluator::
   SetExtraPolynomial(int degree,const ScalarType* extra)
   {
	 
      assert(degree >= -1 && degree <= 2);
    
      m_ExtraPolynomialDegree = degree;

      // set up the extra polynomialTerms

      int numberExtraPolynomialTerms;
      switch(m_ExtraPolynomialDegree)
      {
	 case -1:
	    numberExtraPolynomialTerms = 0;
	    break;
	 case 0:
	    numberExtraPolynomialTerms = 1;
	    break;
	 case 1:
	    numberExtraPolynomialTerms = 4;
	    break;
	 case 2:
	    numberExtraPolynomialTerms = 10;
	    break;
	 default:
	    std::cout << "Error: invalid extra polynomial degree" <<std::endl;
	    abort();
	    break;
      }

      if(degree >= 0 && extra == 0)
      {
	 std::cout << "Error: extra pointer is NULL" <<std::endl;
	 abort();
      }

      m_ExtraPolynomialCoefficients.resize(numberExtraPolynomialTerms);
      for(int i=0;i<numberExtraPolynomialTerms;i++)
	 m_ExtraPolynomialCoefficients[i] = extra[i];

      // Note: m_Modified does not need to be touched here.

   }	 



   // Compute a bounding box of all the points
   void
   evaluator::
   ComputeBoundingBox()
   {


      m_BoundingBox[0] = m_BoundingBox[2] = 
	 m_BoundingBox[4] = std::numeric_limits<ScalarType>::max();
      m_BoundingBox[1] = m_BoundingBox[3] = 
	 m_BoundingBox[5] = std::numeric_limits<ScalarType>::min();


      PointType p;
      
      //iterate over all the points to find 
      // a bounding box of the points
      for(int i=0;i<m_NumberOfPoints;i++)
      {
	 p = m_Points[i];

	 if(p[0] < m_BoundingBox[0])
	   m_BoundingBox[0] = p[0];
	 if(p[0] > m_BoundingBox[1])
	   m_BoundingBox[1] = p[0];

	 if(p[1] < m_BoundingBox[2])
	   m_BoundingBox[2] = p[1];
	 if(p[1] > m_BoundingBox[3])
	   m_BoundingBox[3] = p[1];

	 if(p[2] < m_BoundingBox[4])
	    m_BoundingBox[4] = p[2];
	 if(p[2] > m_BoundingBox[5])
	   m_BoundingBox[5] = p[2];
      }


      const ScalarType eps = 1e-10;
      m_BoundingBox[0] -= eps;
      m_BoundingBox[1] += eps;
      m_BoundingBox[2] -= eps;
      m_BoundingBox[3] += eps;
      m_BoundingBox[4] -= eps;
      m_BoundingBox[5] += eps;


      PRINT(m_BoundingBox[0]);
      PRINT(m_BoundingBox[1]);
      PRINT(m_BoundingBox[2]);
      PRINT(m_BoundingBox[3]);
      PRINT(m_BoundingBox[4]);
      PRINT(m_BoundingBox[5]);

   }


      struct lessthan : public std::binary_function< int, int, bool > 
      {
	    bool operator()(const int& p1, 
			    const int& p2) 
	    { 
	       return pointToBoxMap[p1] < pointToBoxMap[p2];
	    }

	    const int* pointToBoxMap;
      };



   void
   evaluator::SortPoints()
   {

      // We sort the points as follows. We define an array of pointers to each
      // of the points and then we sort this array of pointers using
      // std::sort. Then we rearrange th

      m_OriginalOrderOfPoints.resize(m_NumberOfPoints);

      for(int i=0;i<m_NumberOfPoints;i++)
	 m_OriginalOrderOfPoints[i] = i;

      lessthan lt;
      lt.pointToBoxMap = &m_PointToBoxMap[m_Depth][0];

      std::sort(&m_OriginalOrderOfPoints[0],
		&m_OriginalOrderOfPoints[0]+m_NumberOfPoints,lt);

      // Now reorder m_Points and m_Coefficients and m_PointToBoxMap

      // First make a temporary copy of the points and coefficients and
      // pointsToBoxMap
      const std::vector< PointType > tmpPoints = m_Points;
      const std::valarray< ScalarType > tmpCoef = m_Coefficients;
      std::vector< std::valarray<int> > tmpMap(m_PointToBoxMap.size());

      for(int j=m_Depth;j>=2;j--)
      {
	 tmpMap[j].resize(m_NumberOfPoints);
	 tmpMap[j] = m_PointToBoxMap[j];
      }

      // Now do the actual reordering
      for(int i=0;i<m_NumberOfPoints;++i)
      {
	 const int orig_i = m_OriginalOrderOfPoints[i];

	 m_Points[i] = tmpPoints[orig_i];
	 m_Coefficients[i] = tmpCoef[orig_i];
	 for(int j=m_Depth;j>=2;j--)
	    m_PointToBoxMap[j][i] = tmpMap[j][orig_i];

	 PRINT(m_PointToBoxMap[m_Depth][i]);
      }

/*
      for(int i=0;i<m_NumberOfPoints;i++)
      {
	 const PointType& p = m_Points[i];
	 std::cout << i << ": " << p[0] << " " << p[1] << " " << p[2] << " " 
	           << std::endl;
      }
*/

   }
   

   void
   evaluator::
   Update()
   {

      if(m_Points.size() != m_Coefficients.size())
      {
	 std::cout << "Error. Size of points and coefficients do not match.\n";
	 abort();
      }

      if(m_Modified == false)
      {
	 // If only the coefficients were modified but nothing else, then do
	 // the upward pass and return.
	 if(m_ModifiedCoefficients == true)
	 {
	    // First reorder the coefficients since we've changed the order by
	    // modifying them. 
	    const std::valarray< ScalarType > tmpCoef = m_Coefficients;

	    for(int i=0;i<m_NumberOfPoints;++i)
	    {
	       const int orig_i = m_OriginalOrderOfPoints[i];
	       m_Coefficients[i] = tmpCoef[orig_i];
	    }
	    
	    this->UpwardPass();

	    m_ModifiedCoefficients = false;
	 }

	 return;
      }

      assert(m_NumberOfPoints > 0);
      assert(m_Depth >= 2);
      assert(m_Tau >= 0.0);

      
      this->ComputeBoundingBox();

      const ScalarType dx = m_BoundingBox[1]-m_BoundingBox[0];
      const ScalarType dy = m_BoundingBox[3]-m_BoundingBox[2];
      const ScalarType dz = m_BoundingBox[5]-m_BoundingBox[4];


      // Allocate memory for the far field expansions
      m_FarFieldExpansions.resize(m_Depth+1);
      
      m_CenterPerBox.resize(m_Depth+1);

      m_NumberOfBoxesPerLevel.resize(m_Depth+1);
      m_NumberOfBoxesPerLevel[0] = 1;
      m_NumberOfBoxesPerLevel[1] = 8;

//      m_M.resize(m_Depth+1);
      m_R2.resize(m_Depth+1);
//      m_Error2KMRK.resize(m_Depth+1);
      m_BoxToPointMap.resize(m_Depth+1);
      m_PointToBoxMap.resize(m_Depth+1);


      // Note the first two levels are always empty since they have no
      // polynomials
      for(int i=2;i<=m_Depth;i++)
      {

	 const int numBox = m_NumberOfBoxesPerLevel[i] = 
	    static_cast<int>(std::pow(8.0,i));

	    
	 PolynomialArrayType& polyArray = m_FarFieldExpansions[i];
	 polyArray.resize(numBox);

	 std::vector<PointType>& centerArray = m_CenterPerBox[i];
	 centerArray.resize(numBox);
	    

	 float center[3];

	 // Set degree of polynomials and the centers of each box
	 for(int j=0;j<numBox;j++)
	 {
	    //polyArray[j].SetDegree(m_PplusK);
	    polyArray[j].SetDegreeNoFillZero(m_PplusK);

	    Center(j,i,center);

	    centerArray[j][0] = static_cast<ScalarType>(center[0])*dx + 
	       m_BoundingBox[0];
	    centerArray[j][1] = static_cast<ScalarType>(center[1])*dy + 
	       m_BoundingBox[2];
	    centerArray[j][2] = static_cast<ScalarType>(center[2])*dz +
	       m_BoundingBox[4];

	    //PRINT(centerArray[j][0]);
	    //PRINT(centerArray[j][1]);
	    //PRINT(centerArray[j][2]);
	    DISPLAY(static_cast<ScalarType>(center[0]));
	    DISPLAY(static_cast<ScalarType>(center[1]));
	    DISPLAY(static_cast<ScalarType>(center[2]));
	    DISP("\n");

	 }

//	 m_M[i].resize(numBox);
	 m_R2[i].resize(numBox);
//	 m_Error2KMRK[i].resize(numBox);

//	 m_R2[i] = 0.0;
	 
	 //m_R2[i] = sqrt(std::pow(2.0,-2*(i+1)) + m_Tau2);
	 //m_R2[i] = sqrt(3.0*std::pow(2.0,-2*(i+1)) + m_Tau2);
	 //m_Error2KMRK[i] = m_Error/(std::pow(2.0,m_K)*std::pow(m_R2[i],m_K));
	    
	 m_BoxToPointMap[i].resize(numBox+1);
	 m_PointToBoxMap[i].resize(m_NumberOfPoints);
      }


      // See Eq. 5.4 in paper. We store them here so we don't need to
      // recalculate them. 
      m_A.resize(m_PplusK+1);
      m_B.resize(m_PplusK+1);
      for(int m=2;m<=m_PplusK;m++)
      {
	 m_A[m] = static_cast<ScalarType>(2*m-m_K-2) / 
	    static_cast<ScalarType>(m);
	 m_B[m] = static_cast<ScalarType>(m_K+2-m) / 
	    static_cast<ScalarType>(m);
      }




      ScalarType x,y,z;


      //const ScalarType dx = m_BoundingBox[1]-m_BoundingBox[0];
      //const ScalarType dy = m_BoundingBox[3]-m_BoundingBox[2];
      //const ScalarType dz = m_BoundingBox[5]-m_BoundingBox[4];



      for(int i=0;i<m_NumberOfPoints;i++)
      {



	 x = (m_Points[i][0] - m_BoundingBox[0])/dx;
	 y = (m_Points[i][1] - m_BoundingBox[2])/dy;
	 z = (m_Points[i][2] - m_BoundingBox[4])/dz;



#ifndef NDEBUG
	 if(x <= 0.0 || x >= 1.0 ||
	    y <= 0.0 || y >= 1.0 ||
	    z <= 0.0 || z >= 1.0)
	 {
	    std::cout << "point not within unit cube" << std::endl;
	    abort();
	 }
#endif


	 m_PointToBoxMap[m_Depth][i] = Box(x,y,z,m_Depth);

#ifndef NDEBUG
	 if(m_PointToBoxMap[m_Depth][i] == 0)
	 {
	    PRINT(x << " " << y << " " << z);
	    PRINT(static_cast<float>(x));
	    PRINT(static_cast<float>(y));
	    PRINT(static_cast<float>(z));
	    PRINT(m_Points[i][0]);
	    PRINT(m_Points[i][1]);
	    PRINT(m_Points[i][2]);

	 }
#endif

	 // Set the box of all the ancestors
	 for(int j=m_Depth-1;j>=2;j--)
	    m_PointToBoxMap[j][i] = Parent(m_PointToBoxMap[j+1][i]);
      }




      // If points are unsorted, sort them to speed things up
//      if(m_PointsAreSorted==false)
//      {
	 this->SortPoints();
//	 m_PointsAreSorted = true;
//      }


      // Finally do the upward pass
      this->UpwardPass();


	 
      m_Modified = false;
      m_ModifiedCoefficients = false;

   }


   void
   evaluator::
   UpwardPass()
   {

      PolynomialType G(m_PplusK);   //temporary polynomial

      int nextBox;
      const ScalarType* centerOfBox=0;

      // Note the fourth element holds the magnitude squared
      ScalarType displacement[4];

      //for(int j=2;j>=m_Depth;j++)
      for(int j=m_Depth;j>=2;j--)
      {
	 int currentBox = -1;


	 // Setup the BoxToPointMap in this step. We'll need it later

	 const std::vector<PointType>& centerArray = m_CenterPerBox[j];
	 PolynomialArrayType& polyArray = m_FarFieldExpansions[j];
	 std::valarray<ScalarType>& R2 = m_R2[j];
	 R2 = 0.0; // Set all the radii squared to zero
	 std::valarray<int>& boxToPointMap = m_BoxToPointMap[j];
	 const std::valarray<int>& pointToBoxMap = m_PointToBoxMap[j];
	 
	 for(int i=0;i<m_NumberOfPoints;i++)
	 {


	    // Get the box index of this next point
	    nextBox = pointToBoxMap[i];

	    // If we've entered a new box then get the new center
	    if(currentBox < nextBox)
	    {
	       // do the box to point map
	       boxToPointMap[++currentBox] = i;
	       
	       polyArray[currentBox].fill_zero();
	    
	       while( currentBox < nextBox)
	       {
		  ++currentBox;
		  boxToPointMap[currentBox] = boxToPointMap[currentBox-1];

		  //polyArray[currentBox].fill_zero();
	       }

	       centerOfBox = &centerArray[currentBox][0];
//	    std::cout << centerOfBox[0] << " " << centerOfBox[1] << " " 
//		      << centerOfBox[2] << "\n";
	    
	    }


	    // Now form the p-term multipole expansion

	    // Get the next point
	    const PointType& sourcepoint = m_Points[i];
	    
	    // shift the point so it's relative to the center of the box 
	    displacement[0] = sourcepoint[0] - centerOfBox[0];
	    displacement[1] = sourcepoint[1] - centerOfBox[1];
	    displacement[2] = sourcepoint[2] - centerOfBox[2];

#ifndef NDEBUG
	    if(currentBox ==0 )
	    {
	       PRINT(i);
	       PRINT(j);
	       DISPLAY(sourcepoint[0]);
	       DISPLAY(sourcepoint[1]);
	       DISPLAY(sourcepoint[2]);
	       DISPLAY("\n");
	       DISPLAY(displacement[0]);
	       DISPLAY(displacement[1]);
	       DISPLAY(displacement[2]);
	       DISPLAY("\n");
	       DISPLAY(centerOfBox[0]);
	       DISPLAY(centerOfBox[1]);
	       DISPLAY(centerOfBox[2]);
	       DISPLAY("\n");
	       DISP("\n");
	       
	       
	    }
#endif
		  
	    displacement[3] = (displacement[0]*displacement[0] +
			       displacement[1]*displacement[1] +
			       displacement[2]*displacement[2] +
			       m_Tau2);
	    
	    if(displacement[3] > R2[currentBox])
	       R2[currentBox] = displacement[3];

	    // Generate the coefficients
	    this->PolynomialGenerator(G,displacement,m_Coefficients[i]);

	    // Add these new coefficients to the far field
	    polyArray[currentBox].Add(G);


	 }
      
	 boxToPointMap[++currentBox] = m_NumberOfPoints;
      
	 nextBox = m_NumberOfBoxesPerLevel[j];
	 while( currentBox < nextBox)
	 {
	    ++currentBox;
	    boxToPointMap[currentBox] = boxToPointMap[currentBox-1];
	 }


      }
      
   }





   //inline 
   void
   evaluator::
   PolynomialGenerator(PolynomialType &G, const ScalarType t[4],
		       const ScalarType coef) const
   {

      //const int q = m_PplusK;
      //const int k = m_K;
      //const ScalarType tau2 = m_Tau2;

      // clear the G's
      //G.fill_zero();

      //ScalarType a, b, tmp1,tmp2;
      //ScalarType tmp1,tmp2;

      const ScalarType t0 = t[0];
      const ScalarType t1 = t[1];
      const ScalarType t2 = t[2];
      
      //const ScalarType t0t1t2 = tau2 + t[0]*t[0] + t[1]*t[1] + t[2]*t[2];
      //const ScalarType t0t1t2 = m_Tau2 + t0*t0 + t1*t1 + t2*t2;
      const ScalarType t0t1t2tau2 = t[3];

      G(0) = coef;
      //G(1) = -coef*k*t0;  
      //G(2) = -coef*k*t1;  
      //G(3) = -coef*k*t2;  
      const ScalarType tmp = -coef*m_K;
      G(1) = tmp*t0;  
      G(2) = tmp*t1;  
      G(3) = tmp*t2;  

      const ScalarType* Gptr_minus2 = &G(0); // poly pointer
      const ScalarType* Gptr_minus1 = &G(1); // poly pointer
      ScalarType* Gptr = &G(4);        // poly pointer


      //const int kPlus2 = m_K+2;

      int n = 3;

      // m is used here instead of l as in the paper.
      // Compute the expansion up to degree q.
      for(int m=2; m <= m_PplusK; ++m)
      {

	 //a = -2.0*(k/2.0 - m + 1.0)/m; 
	 //b = ((-m + k + 2.0)/m) * t0t1t2tau2;
//         const ScalarType a = static_cast<ScalarType>(2*m-kPlus2) / 
//	    static_cast<ScalarType>(m);
//	 const ScalarType b =t0t1t2tau2 * static_cast<ScalarType>(kPlus2-m) / 
//	    static_cast<ScalarType>(m);
         const ScalarType a = m_A[m];
	 const ScalarType b = t0t1t2tau2 * m_B[m];

	 

	 int i;

	 n += m + 1;

	 // Zero out the next part of the polynomial
	 //for(ScalarType* ptr = Gptr+n; Gptr != ptr ;)
	 for(i=0; i<n; ++i)
	    *Gptr++ = 0.0;
	 Gptr-=n;

	 for(i=1; i<m; ++i)
	 {
	    const int _2ip1 = 2*i;
	    for(int j=0; j<i; ++j)
	    {
	       
	       const ScalarType tmp1 = (*Gptr_minus1) * a; ++Gptr_minus1;
	       const ScalarType tmp2 = (*Gptr_minus2) * b; ++Gptr_minus2;
	       *Gptr += (tmp1*t0 + tmp2); //Gptr+=i;
	       *(Gptr+i) += tmp1*t1;        //  ++Gptr;
	       ++Gptr;
	       *(Gptr+i) += tmp1*t2;          //Gptr+=i;
	       *(Gptr+_2ip1) += tmp2;             //Gptr+=2;
	       *(Gptr+_2ip1+2) += tmp2;             //Gptr-=(2*i+2);
	       //*Gptr += tmp2;             Gptr-=(_2ip2);
	       //++Gptr;
	    }
	 }
	 
	 for(int j=0; j<i; ++j)
	 {
	    const ScalarType tmp1 = (*Gptr_minus1) * a; ++Gptr_minus1;
	    *Gptr += tmp1*t0;         // Gptr+=i;
	    *(Gptr+i) += tmp1*t1;         // ++Gptr;
	    ++Gptr;
	    *(Gptr+i) += tmp1*t2;         // Gptr-=i; 
	    //++Gptr;
	 }
	 
	 
	 Gptr+=i+1;

      } 
      

   }

   
   

   //inline 
   ScalarType
   evaluator::
   EvaluateExtraPolynomialTerms( const ScalarType thisPoint[3] ) const
   {
      
      
      // Add the polynomial part

      switch(m_ExtraPolynomialDegree)
      {
	 case -1:
	    return 0.0;

	 case 1: 
	    // The monomial basis is:
	    // 1, x, y, z
	 {	 
	    return  
	       m_ExtraPolynomialCoefficients[0]              +
	       m_ExtraPolynomialCoefficients[1]*thisPoint[0] +
	       m_ExtraPolynomialCoefficients[2]*thisPoint[1] +
	       m_ExtraPolynomialCoefficients[3]*thisPoint[2];

	 }
	 
	 //break;
	    
	 case 2:
	    // The monomial basis is:
	    // 1, x, x^2, y, yx, y^2, z, zx, zy, z^2
	 {
	    return 
	       m_ExtraPolynomialCoefficients[0]                           +
	       m_ExtraPolynomialCoefficients[1]*thisPoint[0]              +
	       m_ExtraPolynomialCoefficients[2]*thisPoint[0]*thisPoint[0] +
	       m_ExtraPolynomialCoefficients[3]*thisPoint[1]              +
	       m_ExtraPolynomialCoefficients[4]*thisPoint[1]*thisPoint[0] +
	       m_ExtraPolynomialCoefficients[5]*thisPoint[1]*thisPoint[1] +
	       m_ExtraPolynomialCoefficients[6]*thisPoint[2]              +
	       m_ExtraPolynomialCoefficients[7]*thisPoint[2]*thisPoint[0] +
	       m_ExtraPolynomialCoefficients[8]*thisPoint[2]*thisPoint[1] +
	       m_ExtraPolynomialCoefficients[9]*thisPoint[2]*thisPoint[2];
	 }
	 
	 //break;
	 
	 case 0:
	    // The monomial basis is:
	    // 1
	    
	    return m_ExtraPolynomialCoefficients[0];
	    
	    //break;
	    
	    
      }

      
      return 0.0;

   }



   void
   evaluator::
   EvaluateFarFieldPolynomial( const ScalarType point[4], 
			       const PolynomialType& p,
			       const unsigned int maxDegree,
			       ScalarType& result) const
   {

      
      const ScalarType x = point[0];
      const ScalarType y = point[1];
      const ScalarType z = point[2];


      // Note the fourth element holds the squared magnitude
      const ScalarType x2y2z2   = point[3]; 
      //const ScalarType x2y2z2   = x*x+y*y+z*z; 
      ScalarType       x2y2z2_p = std::sqrt(std::pow(x2y2z2 , -m_K));
      //ScalarType   x2y2z2_p = std::sqrt(std::pow(x2y2z2 , 2*maxDegree - m_K)); 


//      const ScalarType* __restrict const p = &p_[0]; // poly pointer

//      ScalarType result = poly(0) / sqrt(std::pow(x2y2z2 , -m_K));
      
//      const ScalarType* pptr = &poly(1); // poly pointer
//      const ScalarType* p = &poly(0); // poly pointer

      //const ScalarType yDz = y/z; 
      const ScalarType D = y/z; 
//      if(z < 1.0e-6 && z > -1.0e-6)
//	 std::cout << z << " ";
//      ScalarType w;
//      ScalarType* zi = &tmp[0]; // poly pointer
      //ScalarType zi[maxDegree];
      ScalarType zi[21];
//      zi[0] = zi[1] = 0.0;

//0
      result += p[0] / x2y2z2_p;
      if(maxDegree == 0) return;

//1
      //zi[1] = z; 
      x2y2z2_p *= x2y2z2;
      result += (p[1]*x + p[2]*y + p[3]*z) / x2y2z2_p;
      if(maxDegree == 1) return;

// The remainder of this function was automatically generated using the
// program horner.cxx (just compile it with g++ horner.cxx and run ./a.out)
// and should not be edited by hand (except for the very first
// line). Surprisingly, it works pretty well.

//zi[2] = z*zi[1];
zi[2] = z*z;
x2y2z2_p *= x2y2z2;
result += 
(((( p[4] ))*x + 
( (p[5])*D+p[6] )*z)*x + 
( ((p[7])*D+p[8])*D+p[9] )*zi[2]) /
x2y2z2_p;
if(maxDegree == 2) return;


zi[3] = z*zi[2];
x2y2z2_p *= x2y2z2;
result += 
((((( p[10] ))*x + 
( (p[11])*D+p[12] )*z)*x + 
( ((p[13])*D+p[14])*D+p[15] )*zi[2])*x + 
( (((p[16])*D+p[17])*D+p[18])*D+p[19] )*zi[3]) /
x2y2z2_p;
if(maxDegree == 3) return;


zi[4] = z*zi[3];
x2y2z2_p *= x2y2z2;
result += 
(((((( p[20] ))*x + 
( (p[21])*D+p[22] )*z)*x + 
( ((p[23])*D+p[24])*D+p[25] )*zi[2])*x + 
( (((p[26])*D+p[27])*D+p[28])*D+p[29] )*zi[3])*x + 
( ((((p[30])*D+p[31])*D+p[32])*D+p[33])*D+p[34] )*zi[4]) /
x2y2z2_p;
if(maxDegree == 4) return;


zi[5] = z*zi[4];
x2y2z2_p *= x2y2z2;
result += 
((((((( p[35] ))*x + 
( (p[36])*D+p[37] )*z)*x + 
( ((p[38])*D+p[39])*D+p[40] )*zi[2])*x + 
( (((p[41])*D+p[42])*D+p[43])*D+p[44] )*zi[3])*x + 
( ((((p[45])*D+p[46])*D+p[47])*D+p[48])*D+p[49] )*zi[4])*x + 
( (((((p[50])*D+p[51])*D+p[52])*D+p[53])*D+p[54])*D+p[55] )*zi[5]) /
x2y2z2_p;
if(maxDegree == 5) return;


zi[6] = z*zi[5];
x2y2z2_p *= x2y2z2;
result += 
(((((((( p[56] ))*x + 
( (p[57])*D+p[58] )*z)*x + 
( ((p[59])*D+p[60])*D+p[61] )*zi[2])*x + 
( (((p[62])*D+p[63])*D+p[64])*D+p[65] )*zi[3])*x + 
( ((((p[66])*D+p[67])*D+p[68])*D+p[69])*D+p[70] )*zi[4])*x + 
( (((((p[71])*D+p[72])*D+p[73])*D+p[74])*D+p[75])*D+p[76] )*zi[5])*x + 
( ((((((p[77])*D+p[78])*D+p[79])*D+p[80])*D+p[81])*D+p[82])*D+p[83] )*zi[6]) /
x2y2z2_p;
if(maxDegree == 6) return;


zi[7] = z*zi[6];
x2y2z2_p *= x2y2z2;
result += 
((((((((( p[84] ))*x + 
( (p[85])*D+p[86] )*z)*x + 
( ((p[87])*D+p[88])*D+p[89] )*zi[2])*x + 
( (((p[90])*D+p[91])*D+p[92])*D+p[93] )*zi[3])*x + 
( ((((p[94])*D+p[95])*D+p[96])*D+p[97])*D+p[98] )*zi[4])*x + 
( (((((p[99])*D+p[100])*D+p[101])*D+p[102])*D+p[103])*D+p[104] )*zi[5])*x + 
( ((((((p[105])*D+p[106])*D+p[107])*D+p[108])*D+p[109])*D+p[110])*D+p[111] )*zi[6])*x + 
( (((((((p[112])*D+p[113])*D+p[114])*D+p[115])*D+p[116])*D+p[117])*D+p[118])*D+p[119] )*zi[7]) /
x2y2z2_p;
if(maxDegree == 7) return;


zi[8] = z*zi[7];
x2y2z2_p *= x2y2z2;
result += 
(((((((((( p[120] ))*x + 
( (p[121])*D+p[122] )*z)*x + 
( ((p[123])*D+p[124])*D+p[125] )*zi[2])*x + 
( (((p[126])*D+p[127])*D+p[128])*D+p[129] )*zi[3])*x + 
( ((((p[130])*D+p[131])*D+p[132])*D+p[133])*D+p[134] )*zi[4])*x + 
( (((((p[135])*D+p[136])*D+p[137])*D+p[138])*D+p[139])*D+p[140] )*zi[5])*x + 
( ((((((p[141])*D+p[142])*D+p[143])*D+p[144])*D+p[145])*D+p[146])*D+p[147] )*zi[6])*x + 
( (((((((p[148])*D+p[149])*D+p[150])*D+p[151])*D+p[152])*D+p[153])*D+p[154])*D+p[155] )*zi[7])*x + 
( ((((((((p[156])*D+p[157])*D+p[158])*D+p[159])*D+p[160])*D+p[161])*D+p[162])*D+p[163])*D+p[164] )*zi[8]) /
x2y2z2_p;
if(maxDegree == 8) return;


zi[9] = z*zi[8];
x2y2z2_p *= x2y2z2;
result += 
((((((((((( p[165] ))*x + 
( (p[166])*D+p[167] )*z)*x + 
( ((p[168])*D+p[169])*D+p[170] )*zi[2])*x + 
( (((p[171])*D+p[172])*D+p[173])*D+p[174] )*zi[3])*x + 
( ((((p[175])*D+p[176])*D+p[177])*D+p[178])*D+p[179] )*zi[4])*x + 
( (((((p[180])*D+p[181])*D+p[182])*D+p[183])*D+p[184])*D+p[185] )*zi[5])*x + 
( ((((((p[186])*D+p[187])*D+p[188])*D+p[189])*D+p[190])*D+p[191])*D+p[192] )*zi[6])*x + 
( (((((((p[193])*D+p[194])*D+p[195])*D+p[196])*D+p[197])*D+p[198])*D+p[199])*D+p[200] )*zi[7])*x + 
( ((((((((p[201])*D+p[202])*D+p[203])*D+p[204])*D+p[205])*D+p[206])*D+p[207])*D+p[208])*D+p[209] )*zi[8])*x + 
( (((((((((p[210])*D+p[211])*D+p[212])*D+p[213])*D+p[214])*D+p[215])*D+p[216])*D+p[217])*D+p[218])*D+p[219] )*zi[9]) /
x2y2z2_p;
if(maxDegree == 9) return;


zi[10] = z*zi[9];
x2y2z2_p *= x2y2z2;
result += 
(((((((((((( p[220] ))*x + 
( (p[221])*D+p[222] )*z)*x + 
( ((p[223])*D+p[224])*D+p[225] )*zi[2])*x + 
( (((p[226])*D+p[227])*D+p[228])*D+p[229] )*zi[3])*x + 
( ((((p[230])*D+p[231])*D+p[232])*D+p[233])*D+p[234] )*zi[4])*x + 
( (((((p[235])*D+p[236])*D+p[237])*D+p[238])*D+p[239])*D+p[240] )*zi[5])*x + 
( ((((((p[241])*D+p[242])*D+p[243])*D+p[244])*D+p[245])*D+p[246])*D+p[247] )*zi[6])*x + 
( (((((((p[248])*D+p[249])*D+p[250])*D+p[251])*D+p[252])*D+p[253])*D+p[254])*D+p[255] )*zi[7])*x + 
( ((((((((p[256])*D+p[257])*D+p[258])*D+p[259])*D+p[260])*D+p[261])*D+p[262])*D+p[263])*D+p[264] )*zi[8])*x + 
( (((((((((p[265])*D+p[266])*D+p[267])*D+p[268])*D+p[269])*D+p[270])*D+p[271])*D+p[272])*D+p[273])*D+p[274] )*zi[9])*x + 
( ((((((((((p[275])*D+p[276])*D+p[277])*D+p[278])*D+p[279])*D+p[280])*D+p[281])*D+p[282])*D+p[283])*D+p[284])*D+p[285] )*zi[10]) /
x2y2z2_p;
if(maxDegree == 10) return;


zi[11] = z*zi[10];
x2y2z2_p *= x2y2z2;
result += 
((((((((((((( p[286] ))*x + 
( (p[287])*D+p[288] )*z)*x + 
( ((p[289])*D+p[290])*D+p[291] )*zi[2])*x + 
( (((p[292])*D+p[293])*D+p[294])*D+p[295] )*zi[3])*x + 
( ((((p[296])*D+p[297])*D+p[298])*D+p[299])*D+p[300] )*zi[4])*x + 
( (((((p[301])*D+p[302])*D+p[303])*D+p[304])*D+p[305])*D+p[306] )*zi[5])*x + 
( ((((((p[307])*D+p[308])*D+p[309])*D+p[310])*D+p[311])*D+p[312])*D+p[313] )*zi[6])*x + 
( (((((((p[314])*D+p[315])*D+p[316])*D+p[317])*D+p[318])*D+p[319])*D+p[320])*D+p[321] )*zi[7])*x + 
( ((((((((p[322])*D+p[323])*D+p[324])*D+p[325])*D+p[326])*D+p[327])*D+p[328])*D+p[329])*D+p[330] )*zi[8])*x + 
( (((((((((p[331])*D+p[332])*D+p[333])*D+p[334])*D+p[335])*D+p[336])*D+p[337])*D+p[338])*D+p[339])*D+p[340] )*zi[9])*x + 
( ((((((((((p[341])*D+p[342])*D+p[343])*D+p[344])*D+p[345])*D+p[346])*D+p[347])*D+p[348])*D+p[349])*D+p[350])*D+p[351] )*zi[10])*x + 
( (((((((((((p[352])*D+p[353])*D+p[354])*D+p[355])*D+p[356])*D+p[357])*D+p[358])*D+p[359])*D+p[360])*D+p[361])*D+p[362])*D+p[363] )*zi[11]) /
x2y2z2_p;
if(maxDegree == 11) return;


zi[12] = z*zi[11];
x2y2z2_p *= x2y2z2;
result += 
(((((((((((((( p[364] ))*x + 
( (p[365])*D+p[366] )*z)*x + 
( ((p[367])*D+p[368])*D+p[369] )*zi[2])*x + 
( (((p[370])*D+p[371])*D+p[372])*D+p[373] )*zi[3])*x + 
( ((((p[374])*D+p[375])*D+p[376])*D+p[377])*D+p[378] )*zi[4])*x + 
( (((((p[379])*D+p[380])*D+p[381])*D+p[382])*D+p[383])*D+p[384] )*zi[5])*x + 
( ((((((p[385])*D+p[386])*D+p[387])*D+p[388])*D+p[389])*D+p[390])*D+p[391] )*zi[6])*x + 
( (((((((p[392])*D+p[393])*D+p[394])*D+p[395])*D+p[396])*D+p[397])*D+p[398])*D+p[399] )*zi[7])*x + 
( ((((((((p[400])*D+p[401])*D+p[402])*D+p[403])*D+p[404])*D+p[405])*D+p[406])*D+p[407])*D+p[408] )*zi[8])*x + 
( (((((((((p[409])*D+p[410])*D+p[411])*D+p[412])*D+p[413])*D+p[414])*D+p[415])*D+p[416])*D+p[417])*D+p[418] )*zi[9])*x + 
( ((((((((((p[419])*D+p[420])*D+p[421])*D+p[422])*D+p[423])*D+p[424])*D+p[425])*D+p[426])*D+p[427])*D+p[428])*D+p[429] )*zi[10])*x + 
( (((((((((((p[430])*D+p[431])*D+p[432])*D+p[433])*D+p[434])*D+p[435])*D+p[436])*D+p[437])*D+p[438])*D+p[439])*D+p[440])*D+p[441] )*zi[11])*x + 
( ((((((((((((p[442])*D+p[443])*D+p[444])*D+p[445])*D+p[446])*D+p[447])*D+p[448])*D+p[449])*D+p[450])*D+p[451])*D+p[452])*D+p[453])*D+p[454] )*zi[12]) /
x2y2z2_p;
if(maxDegree == 12) return;


zi[13] = z*zi[12];
x2y2z2_p *= x2y2z2;
result += 
((((((((((((((( p[455] ))*x + 
( (p[456])*D+p[457] )*z)*x + 
( ((p[458])*D+p[459])*D+p[460] )*zi[2])*x + 
( (((p[461])*D+p[462])*D+p[463])*D+p[464] )*zi[3])*x + 
( ((((p[465])*D+p[466])*D+p[467])*D+p[468])*D+p[469] )*zi[4])*x + 
( (((((p[470])*D+p[471])*D+p[472])*D+p[473])*D+p[474])*D+p[475] )*zi[5])*x + 
( ((((((p[476])*D+p[477])*D+p[478])*D+p[479])*D+p[480])*D+p[481])*D+p[482] )*zi[6])*x + 
( (((((((p[483])*D+p[484])*D+p[485])*D+p[486])*D+p[487])*D+p[488])*D+p[489])*D+p[490] )*zi[7])*x + 
( ((((((((p[491])*D+p[492])*D+p[493])*D+p[494])*D+p[495])*D+p[496])*D+p[497])*D+p[498])*D+p[499] )*zi[8])*x + 
( (((((((((p[500])*D+p[501])*D+p[502])*D+p[503])*D+p[504])*D+p[505])*D+p[506])*D+p[507])*D+p[508])*D+p[509] )*zi[9])*x + 
( ((((((((((p[510])*D+p[511])*D+p[512])*D+p[513])*D+p[514])*D+p[515])*D+p[516])*D+p[517])*D+p[518])*D+p[519])*D+p[520] )*zi[10])*x + 
( (((((((((((p[521])*D+p[522])*D+p[523])*D+p[524])*D+p[525])*D+p[526])*D+p[527])*D+p[528])*D+p[529])*D+p[530])*D+p[531])*D+p[532] )*zi[11])*x + 
( ((((((((((((p[533])*D+p[534])*D+p[535])*D+p[536])*D+p[537])*D+p[538])*D+p[539])*D+p[540])*D+p[541])*D+p[542])*D+p[543])*D+p[544])*D+p[545] )*zi[12])*x + 
( (((((((((((((p[546])*D+p[547])*D+p[548])*D+p[549])*D+p[550])*D+p[551])*D+p[552])*D+p[553])*D+p[554])*D+p[555])*D+p[556])*D+p[557])*D+p[558])*D+p[559] )*zi[13]) /
x2y2z2_p;
if(maxDegree == 13) return;


zi[14] = z*zi[13];
x2y2z2_p *= x2y2z2;
result += 
(((((((((((((((( p[560] ))*x + 
( (p[561])*D+p[562] )*z)*x + 
( ((p[563])*D+p[564])*D+p[565] )*zi[2])*x + 
( (((p[566])*D+p[567])*D+p[568])*D+p[569] )*zi[3])*x + 
( ((((p[570])*D+p[571])*D+p[572])*D+p[573])*D+p[574] )*zi[4])*x + 
( (((((p[575])*D+p[576])*D+p[577])*D+p[578])*D+p[579])*D+p[580] )*zi[5])*x + 
( ((((((p[581])*D+p[582])*D+p[583])*D+p[584])*D+p[585])*D+p[586])*D+p[587] )*zi[6])*x + 
( (((((((p[588])*D+p[589])*D+p[590])*D+p[591])*D+p[592])*D+p[593])*D+p[594])*D+p[595] )*zi[7])*x + 
( ((((((((p[596])*D+p[597])*D+p[598])*D+p[599])*D+p[600])*D+p[601])*D+p[602])*D+p[603])*D+p[604] )*zi[8])*x + 
( (((((((((p[605])*D+p[606])*D+p[607])*D+p[608])*D+p[609])*D+p[610])*D+p[611])*D+p[612])*D+p[613])*D+p[614] )*zi[9])*x + 
( ((((((((((p[615])*D+p[616])*D+p[617])*D+p[618])*D+p[619])*D+p[620])*D+p[621])*D+p[622])*D+p[623])*D+p[624])*D+p[625] )*zi[10])*x + 
( (((((((((((p[626])*D+p[627])*D+p[628])*D+p[629])*D+p[630])*D+p[631])*D+p[632])*D+p[633])*D+p[634])*D+p[635])*D+p[636])*D+p[637] )*zi[11])*x + 
( ((((((((((((p[638])*D+p[639])*D+p[640])*D+p[641])*D+p[642])*D+p[643])*D+p[644])*D+p[645])*D+p[646])*D+p[647])*D+p[648])*D+p[649])*D+p[650] )*zi[12])*x + 
( (((((((((((((p[651])*D+p[652])*D+p[653])*D+p[654])*D+p[655])*D+p[656])*D+p[657])*D+p[658])*D+p[659])*D+p[660])*D+p[661])*D+p[662])*D+p[663])*D+p[664] )*zi[13])*x + 
( ((((((((((((((p[665])*D+p[666])*D+p[667])*D+p[668])*D+p[669])*D+p[670])*D+p[671])*D+p[672])*D+p[673])*D+p[674])*D+p[675])*D+p[676])*D+p[677])*D+p[678])*D+p[679] )*zi[14]) /
x2y2z2_p;
if(maxDegree == 14) return;


zi[15] = z*zi[14];
x2y2z2_p *= x2y2z2;
result += 
((((((((((((((((( p[680] ))*x + 
( (p[681])*D+p[682] )*z)*x + 
( ((p[683])*D+p[684])*D+p[685] )*zi[2])*x + 
( (((p[686])*D+p[687])*D+p[688])*D+p[689] )*zi[3])*x + 
( ((((p[690])*D+p[691])*D+p[692])*D+p[693])*D+p[694] )*zi[4])*x + 
( (((((p[695])*D+p[696])*D+p[697])*D+p[698])*D+p[699])*D+p[700] )*zi[5])*x + 
( ((((((p[701])*D+p[702])*D+p[703])*D+p[704])*D+p[705])*D+p[706])*D+p[707] )*zi[6])*x + 
( (((((((p[708])*D+p[709])*D+p[710])*D+p[711])*D+p[712])*D+p[713])*D+p[714])*D+p[715] )*zi[7])*x + 
( ((((((((p[716])*D+p[717])*D+p[718])*D+p[719])*D+p[720])*D+p[721])*D+p[722])*D+p[723])*D+p[724] )*zi[8])*x + 
( (((((((((p[725])*D+p[726])*D+p[727])*D+p[728])*D+p[729])*D+p[730])*D+p[731])*D+p[732])*D+p[733])*D+p[734] )*zi[9])*x + 
( ((((((((((p[735])*D+p[736])*D+p[737])*D+p[738])*D+p[739])*D+p[740])*D+p[741])*D+p[742])*D+p[743])*D+p[744])*D+p[745] )*zi[10])*x + 
( (((((((((((p[746])*D+p[747])*D+p[748])*D+p[749])*D+p[750])*D+p[751])*D+p[752])*D+p[753])*D+p[754])*D+p[755])*D+p[756])*D+p[757] )*zi[11])*x + 
( ((((((((((((p[758])*D+p[759])*D+p[760])*D+p[761])*D+p[762])*D+p[763])*D+p[764])*D+p[765])*D+p[766])*D+p[767])*D+p[768])*D+p[769])*D+p[770] )*zi[12])*x + 
( (((((((((((((p[771])*D+p[772])*D+p[773])*D+p[774])*D+p[775])*D+p[776])*D+p[777])*D+p[778])*D+p[779])*D+p[780])*D+p[781])*D+p[782])*D+p[783])*D+p[784] )*zi[13])*x + 
( ((((((((((((((p[785])*D+p[786])*D+p[787])*D+p[788])*D+p[789])*D+p[790])*D+p[791])*D+p[792])*D+p[793])*D+p[794])*D+p[795])*D+p[796])*D+p[797])*D+p[798])*D+p[799] )*zi[14])*x + 
( (((((((((((((((p[800])*D+p[801])*D+p[802])*D+p[803])*D+p[804])*D+p[805])*D+p[806])*D+p[807])*D+p[808])*D+p[809])*D+p[810])*D+p[811])*D+p[812])*D+p[813])*D+p[814])*D+p[815] )*zi[15]) /
x2y2z2_p;
if(maxDegree == 15) return;


zi[16] = z*zi[15];
x2y2z2_p *= x2y2z2;
result += 
(((((((((((((((((( p[816] ))*x + 
( (p[817])*D+p[818] )*z)*x + 
( ((p[819])*D+p[820])*D+p[821] )*zi[2])*x + 
( (((p[822])*D+p[823])*D+p[824])*D+p[825] )*zi[3])*x + 
( ((((p[826])*D+p[827])*D+p[828])*D+p[829])*D+p[830] )*zi[4])*x + 
( (((((p[831])*D+p[832])*D+p[833])*D+p[834])*D+p[835])*D+p[836] )*zi[5])*x + 
( ((((((p[837])*D+p[838])*D+p[839])*D+p[840])*D+p[841])*D+p[842])*D+p[843] )*zi[6])*x + 
( (((((((p[844])*D+p[845])*D+p[846])*D+p[847])*D+p[848])*D+p[849])*D+p[850])*D+p[851] )*zi[7])*x + 
( ((((((((p[852])*D+p[853])*D+p[854])*D+p[855])*D+p[856])*D+p[857])*D+p[858])*D+p[859])*D+p[860] )*zi[8])*x + 
( (((((((((p[861])*D+p[862])*D+p[863])*D+p[864])*D+p[865])*D+p[866])*D+p[867])*D+p[868])*D+p[869])*D+p[870] )*zi[9])*x + 
( ((((((((((p[871])*D+p[872])*D+p[873])*D+p[874])*D+p[875])*D+p[876])*D+p[877])*D+p[878])*D+p[879])*D+p[880])*D+p[881] )*zi[10])*x + 
( (((((((((((p[882])*D+p[883])*D+p[884])*D+p[885])*D+p[886])*D+p[887])*D+p[888])*D+p[889])*D+p[890])*D+p[891])*D+p[892])*D+p[893] )*zi[11])*x + 
( ((((((((((((p[894])*D+p[895])*D+p[896])*D+p[897])*D+p[898])*D+p[899])*D+p[900])*D+p[901])*D+p[902])*D+p[903])*D+p[904])*D+p[905])*D+p[906] )*zi[12])*x + 
( (((((((((((((p[907])*D+p[908])*D+p[909])*D+p[910])*D+p[911])*D+p[912])*D+p[913])*D+p[914])*D+p[915])*D+p[916])*D+p[917])*D+p[918])*D+p[919])*D+p[920] )*zi[13])*x + 
( ((((((((((((((p[921])*D+p[922])*D+p[923])*D+p[924])*D+p[925])*D+p[926])*D+p[927])*D+p[928])*D+p[929])*D+p[930])*D+p[931])*D+p[932])*D+p[933])*D+p[934])*D+p[935] )*zi[14])*x + 
( (((((((((((((((p[936])*D+p[937])*D+p[938])*D+p[939])*D+p[940])*D+p[941])*D+p[942])*D+p[943])*D+p[944])*D+p[945])*D+p[946])*D+p[947])*D+p[948])*D+p[949])*D+p[950])*D+p[951] )*zi[15])*x + 
( ((((((((((((((((p[952])*D+p[953])*D+p[954])*D+p[955])*D+p[956])*D+p[957])*D+p[958])*D+p[959])*D+p[960])*D+p[961])*D+p[962])*D+p[963])*D+p[964])*D+p[965])*D+p[966])*D+p[967])*D+p[968] )*zi[16]) /
x2y2z2_p;
if(maxDegree == 16) return;


zi[17] = z*zi[16];
x2y2z2_p *= x2y2z2;
result += 
((((((((((((((((((( p[969] ))*x + 
( (p[970])*D+p[971] )*z)*x + 
( ((p[972])*D+p[973])*D+p[974] )*zi[2])*x + 
( (((p[975])*D+p[976])*D+p[977])*D+p[978] )*zi[3])*x + 
( ((((p[979])*D+p[980])*D+p[981])*D+p[982])*D+p[983] )*zi[4])*x + 
( (((((p[984])*D+p[985])*D+p[986])*D+p[987])*D+p[988])*D+p[989] )*zi[5])*x + 
( ((((((p[990])*D+p[991])*D+p[992])*D+p[993])*D+p[994])*D+p[995])*D+p[996] )*zi[6])*x + 
( (((((((p[997])*D+p[998])*D+p[999])*D+p[1000])*D+p[1001])*D+p[1002])*D+p[1003])*D+p[1004] )*zi[7])*x + 
( ((((((((p[1005])*D+p[1006])*D+p[1007])*D+p[1008])*D+p[1009])*D+p[1010])*D+p[1011])*D+p[1012])*D+p[1013] )*zi[8])*x + 
( (((((((((p[1014])*D+p[1015])*D+p[1016])*D+p[1017])*D+p[1018])*D+p[1019])*D+p[1020])*D+p[1021])*D+p[1022])*D+p[1023] )*zi[9])*x + 
( ((((((((((p[1024])*D+p[1025])*D+p[1026])*D+p[1027])*D+p[1028])*D+p[1029])*D+p[1030])*D+p[1031])*D+p[1032])*D+p[1033])*D+p[1034] )*zi[10])*x + 
( (((((((((((p[1035])*D+p[1036])*D+p[1037])*D+p[1038])*D+p[1039])*D+p[1040])*D+p[1041])*D+p[1042])*D+p[1043])*D+p[1044])*D+p[1045])*D+p[1046] )*zi[11])*x + 
( ((((((((((((p[1047])*D+p[1048])*D+p[1049])*D+p[1050])*D+p[1051])*D+p[1052])*D+p[1053])*D+p[1054])*D+p[1055])*D+p[1056])*D+p[1057])*D+p[1058])*D+p[1059] )*zi[12])*x + 
( (((((((((((((p[1060])*D+p[1061])*D+p[1062])*D+p[1063])*D+p[1064])*D+p[1065])*D+p[1066])*D+p[1067])*D+p[1068])*D+p[1069])*D+p[1070])*D+p[1071])*D+p[1072])*D+p[1073] )*zi[13])*x + 
( ((((((((((((((p[1074])*D+p[1075])*D+p[1076])*D+p[1077])*D+p[1078])*D+p[1079])*D+p[1080])*D+p[1081])*D+p[1082])*D+p[1083])*D+p[1084])*D+p[1085])*D+p[1086])*D+p[1087])*D+p[1088] )*zi[14])*x + 
( (((((((((((((((p[1089])*D+p[1090])*D+p[1091])*D+p[1092])*D+p[1093])*D+p[1094])*D+p[1095])*D+p[1096])*D+p[1097])*D+p[1098])*D+p[1099])*D+p[1100])*D+p[1101])*D+p[1102])*D+p[1103])*D+p[1104] )*zi[15])*x + 
( ((((((((((((((((p[1105])*D+p[1106])*D+p[1107])*D+p[1108])*D+p[1109])*D+p[1110])*D+p[1111])*D+p[1112])*D+p[1113])*D+p[1114])*D+p[1115])*D+p[1116])*D+p[1117])*D+p[1118])*D+p[1119])*D+p[1120])*D+p[1121] )*zi[16])*x + 
( (((((((((((((((((p[1122])*D+p[1123])*D+p[1124])*D+p[1125])*D+p[1126])*D+p[1127])*D+p[1128])*D+p[1129])*D+p[1130])*D+p[1131])*D+p[1132])*D+p[1133])*D+p[1134])*D+p[1135])*D+p[1136])*D+p[1137])*D+p[1138])*D+p[1139] )*zi[17]) /
x2y2z2_p;
if(maxDegree == 17) return;


zi[18] = z*zi[17];
x2y2z2_p *= x2y2z2;
result += 
(((((((((((((((((((( p[1140] ))*x + 
( (p[1141])*D+p[1142] )*z)*x + 
( ((p[1143])*D+p[1144])*D+p[1145] )*zi[2])*x + 
( (((p[1146])*D+p[1147])*D+p[1148])*D+p[1149] )*zi[3])*x + 
( ((((p[1150])*D+p[1151])*D+p[1152])*D+p[1153])*D+p[1154] )*zi[4])*x + 
( (((((p[1155])*D+p[1156])*D+p[1157])*D+p[1158])*D+p[1159])*D+p[1160] )*zi[5])*x + 
( ((((((p[1161])*D+p[1162])*D+p[1163])*D+p[1164])*D+p[1165])*D+p[1166])*D+p[1167] )*zi[6])*x + 
( (((((((p[1168])*D+p[1169])*D+p[1170])*D+p[1171])*D+p[1172])*D+p[1173])*D+p[1174])*D+p[1175] )*zi[7])*x + 
( ((((((((p[1176])*D+p[1177])*D+p[1178])*D+p[1179])*D+p[1180])*D+p[1181])*D+p[1182])*D+p[1183])*D+p[1184] )*zi[8])*x + 
( (((((((((p[1185])*D+p[1186])*D+p[1187])*D+p[1188])*D+p[1189])*D+p[1190])*D+p[1191])*D+p[1192])*D+p[1193])*D+p[1194] )*zi[9])*x + 
( ((((((((((p[1195])*D+p[1196])*D+p[1197])*D+p[1198])*D+p[1199])*D+p[1200])*D+p[1201])*D+p[1202])*D+p[1203])*D+p[1204])*D+p[1205] )*zi[10])*x + 
( (((((((((((p[1206])*D+p[1207])*D+p[1208])*D+p[1209])*D+p[1210])*D+p[1211])*D+p[1212])*D+p[1213])*D+p[1214])*D+p[1215])*D+p[1216])*D+p[1217] )*zi[11])*x + 
( ((((((((((((p[1218])*D+p[1219])*D+p[1220])*D+p[1221])*D+p[1222])*D+p[1223])*D+p[1224])*D+p[1225])*D+p[1226])*D+p[1227])*D+p[1228])*D+p[1229])*D+p[1230] )*zi[12])*x + 
( (((((((((((((p[1231])*D+p[1232])*D+p[1233])*D+p[1234])*D+p[1235])*D+p[1236])*D+p[1237])*D+p[1238])*D+p[1239])*D+p[1240])*D+p[1241])*D+p[1242])*D+p[1243])*D+p[1244] )*zi[13])*x + 
( ((((((((((((((p[1245])*D+p[1246])*D+p[1247])*D+p[1248])*D+p[1249])*D+p[1250])*D+p[1251])*D+p[1252])*D+p[1253])*D+p[1254])*D+p[1255])*D+p[1256])*D+p[1257])*D+p[1258])*D+p[1259] )*zi[14])*x + 
( (((((((((((((((p[1260])*D+p[1261])*D+p[1262])*D+p[1263])*D+p[1264])*D+p[1265])*D+p[1266])*D+p[1267])*D+p[1268])*D+p[1269])*D+p[1270])*D+p[1271])*D+p[1272])*D+p[1273])*D+p[1274])*D+p[1275] )*zi[15])*x + 
( ((((((((((((((((p[1276])*D+p[1277])*D+p[1278])*D+p[1279])*D+p[1280])*D+p[1281])*D+p[1282])*D+p[1283])*D+p[1284])*D+p[1285])*D+p[1286])*D+p[1287])*D+p[1288])*D+p[1289])*D+p[1290])*D+p[1291])*D+p[1292] )*zi[16])*x + 
( (((((((((((((((((p[1293])*D+p[1294])*D+p[1295])*D+p[1296])*D+p[1297])*D+p[1298])*D+p[1299])*D+p[1300])*D+p[1301])*D+p[1302])*D+p[1303])*D+p[1304])*D+p[1305])*D+p[1306])*D+p[1307])*D+p[1308])*D+p[1309])*D+p[1310] )*zi[17])*x + 
( ((((((((((((((((((p[1311])*D+p[1312])*D+p[1313])*D+p[1314])*D+p[1315])*D+p[1316])*D+p[1317])*D+p[1318])*D+p[1319])*D+p[1320])*D+p[1321])*D+p[1322])*D+p[1323])*D+p[1324])*D+p[1325])*D+p[1326])*D+p[1327])*D+p[1328])*D+p[1329] )*zi[18]) /
x2y2z2_p;
if(maxDegree == 18) return;


zi[19] = z*zi[18];
x2y2z2_p *= x2y2z2;
result += 
((((((((((((((((((((( p[1330] ))*x + 
( (p[1331])*D+p[1332] )*z)*x + 
( ((p[1333])*D+p[1334])*D+p[1335] )*zi[2])*x + 
( (((p[1336])*D+p[1337])*D+p[1338])*D+p[1339] )*zi[3])*x + 
( ((((p[1340])*D+p[1341])*D+p[1342])*D+p[1343])*D+p[1344] )*zi[4])*x + 
( (((((p[1345])*D+p[1346])*D+p[1347])*D+p[1348])*D+p[1349])*D+p[1350] )*zi[5])*x + 
( ((((((p[1351])*D+p[1352])*D+p[1353])*D+p[1354])*D+p[1355])*D+p[1356])*D+p[1357] )*zi[6])*x + 
( (((((((p[1358])*D+p[1359])*D+p[1360])*D+p[1361])*D+p[1362])*D+p[1363])*D+p[1364])*D+p[1365] )*zi[7])*x + 
( ((((((((p[1366])*D+p[1367])*D+p[1368])*D+p[1369])*D+p[1370])*D+p[1371])*D+p[1372])*D+p[1373])*D+p[1374] )*zi[8])*x + 
( (((((((((p[1375])*D+p[1376])*D+p[1377])*D+p[1378])*D+p[1379])*D+p[1380])*D+p[1381])*D+p[1382])*D+p[1383])*D+p[1384] )*zi[9])*x + 
( ((((((((((p[1385])*D+p[1386])*D+p[1387])*D+p[1388])*D+p[1389])*D+p[1390])*D+p[1391])*D+p[1392])*D+p[1393])*D+p[1394])*D+p[1395] )*zi[10])*x + 
( (((((((((((p[1396])*D+p[1397])*D+p[1398])*D+p[1399])*D+p[1400])*D+p[1401])*D+p[1402])*D+p[1403])*D+p[1404])*D+p[1405])*D+p[1406])*D+p[1407] )*zi[11])*x + 
( ((((((((((((p[1408])*D+p[1409])*D+p[1410])*D+p[1411])*D+p[1412])*D+p[1413])*D+p[1414])*D+p[1415])*D+p[1416])*D+p[1417])*D+p[1418])*D+p[1419])*D+p[1420] )*zi[12])*x + 
( (((((((((((((p[1421])*D+p[1422])*D+p[1423])*D+p[1424])*D+p[1425])*D+p[1426])*D+p[1427])*D+p[1428])*D+p[1429])*D+p[1430])*D+p[1431])*D+p[1432])*D+p[1433])*D+p[1434] )*zi[13])*x + 
( ((((((((((((((p[1435])*D+p[1436])*D+p[1437])*D+p[1438])*D+p[1439])*D+p[1440])*D+p[1441])*D+p[1442])*D+p[1443])*D+p[1444])*D+p[1445])*D+p[1446])*D+p[1447])*D+p[1448])*D+p[1449] )*zi[14])*x + 
( (((((((((((((((p[1450])*D+p[1451])*D+p[1452])*D+p[1453])*D+p[1454])*D+p[1455])*D+p[1456])*D+p[1457])*D+p[1458])*D+p[1459])*D+p[1460])*D+p[1461])*D+p[1462])*D+p[1463])*D+p[1464])*D+p[1465] )*zi[15])*x + 
( ((((((((((((((((p[1466])*D+p[1467])*D+p[1468])*D+p[1469])*D+p[1470])*D+p[1471])*D+p[1472])*D+p[1473])*D+p[1474])*D+p[1475])*D+p[1476])*D+p[1477])*D+p[1478])*D+p[1479])*D+p[1480])*D+p[1481])*D+p[1482] )*zi[16])*x + 
( (((((((((((((((((p[1483])*D+p[1484])*D+p[1485])*D+p[1486])*D+p[1487])*D+p[1488])*D+p[1489])*D+p[1490])*D+p[1491])*D+p[1492])*D+p[1493])*D+p[1494])*D+p[1495])*D+p[1496])*D+p[1497])*D+p[1498])*D+p[1499])*D+p[1500] )*zi[17])*x + 
( ((((((((((((((((((p[1501])*D+p[1502])*D+p[1503])*D+p[1504])*D+p[1505])*D+p[1506])*D+p[1507])*D+p[1508])*D+p[1509])*D+p[1510])*D+p[1511])*D+p[1512])*D+p[1513])*D+p[1514])*D+p[1515])*D+p[1516])*D+p[1517])*D+p[1518])*D+p[1519] )*zi[18])*x + 
( (((((((((((((((((((p[1520])*D+p[1521])*D+p[1522])*D+p[1523])*D+p[1524])*D+p[1525])*D+p[1526])*D+p[1527])*D+p[1528])*D+p[1529])*D+p[1530])*D+p[1531])*D+p[1532])*D+p[1533])*D+p[1534])*D+p[1535])*D+p[1536])*D+p[1537])*D+p[1538])*D+p[1539] )*zi[19]) /
x2y2z2_p;
if(maxDegree == 19) return;


zi[20] = z*zi[19];
x2y2z2_p *= x2y2z2;
result += 
(((((((((((((((((((((( p[1540] ))*x + 
( (p[1541])*D+p[1542] )*z)*x + 
( ((p[1543])*D+p[1544])*D+p[1545] )*zi[2])*x + 
( (((p[1546])*D+p[1547])*D+p[1548])*D+p[1549] )*zi[3])*x + 
( ((((p[1550])*D+p[1551])*D+p[1552])*D+p[1553])*D+p[1554] )*zi[4])*x + 
( (((((p[1555])*D+p[1556])*D+p[1557])*D+p[1558])*D+p[1559])*D+p[1560] )*zi[5])*x + 
( ((((((p[1561])*D+p[1562])*D+p[1563])*D+p[1564])*D+p[1565])*D+p[1566])*D+p[1567] )*zi[6])*x + 
( (((((((p[1568])*D+p[1569])*D+p[1570])*D+p[1571])*D+p[1572])*D+p[1573])*D+p[1574])*D+p[1575] )*zi[7])*x + 
( ((((((((p[1576])*D+p[1577])*D+p[1578])*D+p[1579])*D+p[1580])*D+p[1581])*D+p[1582])*D+p[1583])*D+p[1584] )*zi[8])*x + 
( (((((((((p[1585])*D+p[1586])*D+p[1587])*D+p[1588])*D+p[1589])*D+p[1590])*D+p[1591])*D+p[1592])*D+p[1593])*D+p[1594] )*zi[9])*x + 
( ((((((((((p[1595])*D+p[1596])*D+p[1597])*D+p[1598])*D+p[1599])*D+p[1600])*D+p[1601])*D+p[1602])*D+p[1603])*D+p[1604])*D+p[1605] )*zi[10])*x + 
( (((((((((((p[1606])*D+p[1607])*D+p[1608])*D+p[1609])*D+p[1610])*D+p[1611])*D+p[1612])*D+p[1613])*D+p[1614])*D+p[1615])*D+p[1616])*D+p[1617] )*zi[11])*x + 
( ((((((((((((p[1618])*D+p[1619])*D+p[1620])*D+p[1621])*D+p[1622])*D+p[1623])*D+p[1624])*D+p[1625])*D+p[1626])*D+p[1627])*D+p[1628])*D+p[1629])*D+p[1630] )*zi[12])*x + 
( (((((((((((((p[1631])*D+p[1632])*D+p[1633])*D+p[1634])*D+p[1635])*D+p[1636])*D+p[1637])*D+p[1638])*D+p[1639])*D+p[1640])*D+p[1641])*D+p[1642])*D+p[1643])*D+p[1644] )*zi[13])*x + 
( ((((((((((((((p[1645])*D+p[1646])*D+p[1647])*D+p[1648])*D+p[1649])*D+p[1650])*D+p[1651])*D+p[1652])*D+p[1653])*D+p[1654])*D+p[1655])*D+p[1656])*D+p[1657])*D+p[1658])*D+p[1659] )*zi[14])*x + 
( (((((((((((((((p[1660])*D+p[1661])*D+p[1662])*D+p[1663])*D+p[1664])*D+p[1665])*D+p[1666])*D+p[1667])*D+p[1668])*D+p[1669])*D+p[1670])*D+p[1671])*D+p[1672])*D+p[1673])*D+p[1674])*D+p[1675] )*zi[15])*x + 
( ((((((((((((((((p[1676])*D+p[1677])*D+p[1678])*D+p[1679])*D+p[1680])*D+p[1681])*D+p[1682])*D+p[1683])*D+p[1684])*D+p[1685])*D+p[1686])*D+p[1687])*D+p[1688])*D+p[1689])*D+p[1690])*D+p[1691])*D+p[1692] )*zi[16])*x + 
( (((((((((((((((((p[1693])*D+p[1694])*D+p[1695])*D+p[1696])*D+p[1697])*D+p[1698])*D+p[1699])*D+p[1700])*D+p[1701])*D+p[1702])*D+p[1703])*D+p[1704])*D+p[1705])*D+p[1706])*D+p[1707])*D+p[1708])*D+p[1709])*D+p[1710] )*zi[17])*x + 
( ((((((((((((((((((p[1711])*D+p[1712])*D+p[1713])*D+p[1714])*D+p[1715])*D+p[1716])*D+p[1717])*D+p[1718])*D+p[1719])*D+p[1720])*D+p[1721])*D+p[1722])*D+p[1723])*D+p[1724])*D+p[1725])*D+p[1726])*D+p[1727])*D+p[1728])*D+p[1729] )*zi[18])*x + 
( (((((((((((((((((((p[1730])*D+p[1731])*D+p[1732])*D+p[1733])*D+p[1734])*D+p[1735])*D+p[1736])*D+p[1737])*D+p[1738])*D+p[1739])*D+p[1740])*D+p[1741])*D+p[1742])*D+p[1743])*D+p[1744])*D+p[1745])*D+p[1746])*D+p[1747])*D+p[1748])*D+p[1749] )*zi[19])*x + 
( ((((((((((((((((((((p[1750])*D+p[1751])*D+p[1752])*D+p[1753])*D+p[1754])*D+p[1755])*D+p[1756])*D+p[1757])*D+p[1758])*D+p[1759])*D+p[1760])*D+p[1761])*D+p[1762])*D+p[1763])*D+p[1764])*D+p[1765])*D+p[1766])*D+p[1767])*D+p[1768])*D+p[1769])*D+p[1770] )*zi[20]) /
x2y2z2_p;
if(maxDegree == 20) return;





   }




   void
   evaluator::
   EvaluateAtAllSourcePoints(ScalarType* const orderedResult) const
   {


      int degree;

      // Note the 4th element holds the magnitude
      ScalarType displacement[4];


      //std::valarray<ScalarType> result(m_NumberOfPoints);
      ScalarType result[m_NumberOfPoints];


      std::vector< std::pair<int,int> > openPanels1(m_NumberOfBoxesPerLevel[m_Depth-1]);
      std::vector< std::pair<int,int> > openPanels2(m_NumberOfBoxesPerLevel[m_Depth-1]);
      std::vector< std::pair<int,int> >* nextOpenPanels;
      std::vector< std::pair<int,int> >* currentOpenPanels;
      int currentOpenPanelsSize;
      int nextOpenPanelsSize;

      enum { EVALUATE_FAR_FIELD,
	     EVALUATE_DIRECT,
	     DEFER_TO_NEXT_LEVEL };

      unsigned int whatToDo;

      for(int i=0;i<m_NumberOfPoints;++i)
      {

	 const ScalarType* const  sourcepoint = &m_Points[i][0];
	 

//------------------------------------------------
	 // Evaluate extra polynomial term

	 result[i] = this->EvaluateExtraPolynomialTerms(sourcepoint);
	 
	 //std::cout << result[i] << " ";


//------------------------------------------------


	 openPanels1[0] = std::make_pair(0,64);
	 currentOpenPanelsSize = 1;

	 currentOpenPanels = &openPanels1;
	 nextOpenPanels = &openPanels2;

	 // do big loop
	 for(int k=2;
	     k<=m_Depth;
	     ++k,currentOpenPanelsSize=nextOpenPanelsSize,
		std::swap(currentOpenPanels,nextOpenPanels))
	 {

	    nextOpenPanelsSize = 0;

	    const std::vector<PointType>& centerArray = m_CenterPerBox[k];
	    const PolynomialArrayType& polyArray = m_FarFieldExpansions[k];
	    //const std::valarray<ScalarType>& M = m_M[k];
	    const std::valarray<ScalarType>& R2 = m_R2[k];
	    //const std::valarray<ScalarType>& E = m_Error2KMRK[k];
	    const std::valarray<int>& boxToPointMap = m_BoxToPointMap[k];


	    for(int jj=0;jj<currentOpenPanelsSize;++jj)
	    {

	       const int second = (*currentOpenPanels)[jj].second;

	       for(int j=(*currentOpenPanels)[jj].first; j<second; ++j)
	       {
	       

		  const PointType& centerOfBox = centerArray[j];
		     
		  // shift the point so it's relative to the center of the
		  // box
		  displacement[0] = sourcepoint[0] - centerOfBox[0];
		  displacement[1] = sourcepoint[1] - centerOfBox[1];
		  displacement[2] = sourcepoint[2] - centerOfBox[2];

	       
		  // Determine if we should evaluate the far field or the
		  // basis functions directly to compute the influence from
		  // this panel. The variable UseFarField will be set to
		  // true if we should use the far field, otherwise it will
		  // be set to false.  Also, if UseFarField is set to true,
		  // determine to what degree we should use for the
		  // evaluation


		  displacement[3] = (displacement[0]*displacement[0]+
				     displacement[1]*displacement[1]+
				     displacement[2]*displacement[2]);

		  //const ScalarType distance = sqrt(displacement[3]);
		  const ScalarType distance2 = displacement[3];


		  // Given a certain desired accuracy and the distance form
		  // the panel center, we now need to compute the degree of
		  // polynomial that will give us the desired accuracy.

		  // If we're too close then don't evaluate far field.
		  if(distance2 <= m_WellSeparatedFactor2*R2[j])
		  {
		     if(k == m_Depth)
		     {
			// If we're at the lowest level, just evaluate direct
			whatToDo = EVALUATE_DIRECT;
		     }
		     else
		     {
			whatToDo = DEFER_TO_NEXT_LEVEL;
		     }
		  }
		  else
		  {

		     // Divide the distance away from the point by the
		     // radius. Subtract this value from m_PplusK

		     //degree = m_PplusK - static_cast<int>(m_DecreaseFactor *
		     //				  sqrt(distance2/R2[j]));
		     if(R2[j] > 1e-8)
			degree =  	m_PplusK - static_cast<int>(
			   m_DecreaseFactor * sqrt(distance2/R2[j]) +
			   m_DecreaseFactor2);
		     else
			degree =  	m_PplusK - static_cast<int>(
			   /*m_DecreaseFactor * sqrt(distance2/R2[j]) +*/
			   m_DecreaseFactor2);
			

		     // clamp to min degree
		     if(degree < m_MinDegree)
			degree = m_MinDegree;
		     // clamp to max degree
		     else if(degree > m_PplusK)
			degree = m_PplusK;

		     // std::cout << degree << std::endl;

		     // Now that we know the degree we must choose among 3
		     // options. If the number of points in the box is less than
		     // the cutoff, simply evaluate directly. If not, then if the
		     // degree is less than m_PplusK then compute the far
		     // field. If the degree is equal to m_PlusK then (since the
		     // far field cannot provide us with a good approximation) if
		     // we are at the lowest level just evaluate
		     // direct. Otherwise defer to the next level.
		     const int numberOfPoints = 
			boxToPointMap[j+1] - boxToPointMap[j];

		     //if(0)
		     if(numberOfPoints < m_NumberCutoffPoints[degree])
		     {
			whatToDo = EVALUATE_DIRECT;//PRINTND(numberOfPoints<<" "<<degree);
		     }
		     else //if(degree <= m_PplusK+1)
		     {
			whatToDo = EVALUATE_FAR_FIELD;
		     }
		  }

		  switch(whatToDo)
		  {
		     case EVALUATE_FAR_FIELD:
		     {
			//std::cout << EVALUATE_FAR_FIELD;
			this->EvaluateFarFieldPolynomial(
			   displacement, polyArray[j], degree, result[i]);
			break;
		     }
		     case EVALUATE_DIRECT:
		     {
			
			//std::cout << EVALUATE_DIRECT;
			this->EvaluateSumBasicFunctionsNew(sourcepoint,
							   &boxToPointMap[j],
							   result[i]);
			break;
		     }
		     case DEFER_TO_NEXT_LEVEL:
		     {
			//std::cout << DEFER_TO_NEXT_LEVEL;
			const int tmp = ChildFirst(j);
			(*nextOpenPanels)[nextOpenPanelsSize++] = 
			   std::make_pair(tmp,tmp+8);
			break;
		     }
		     /* default:
		     {
			abort();
			} */
		  }

	       }
	    }
	 } // end big loop
      } // end for m_NumberOfPoints


      // Now reorder result so it has the original order of the points
      for(int i=0;i<m_NumberOfPoints;++i)
      {
//	 orderedResult[i] = result[ m_OriginalOrderOfPoints[i] ];
	 orderedResult[ m_OriginalOrderOfPoints[i] ] = result[i];
      }

   }//end function



   // Evaluate using FMM at a single point
   ScalarType
   evaluator::
   Evaluate( const ScalarType sourcepoint[3] ) const
   {


      int degree;

      // Note the 4th element holds the magnitude
      ScalarType displacement[4];


      //std::valarray<ScalarType> result(m_NumberOfPoints);
      //ScalarType result[m_NumberOfPoints];
      ScalarType result=0.0;


      std::vector< std::pair<int,int> > openPanels1(m_NumberOfBoxesPerLevel[m_Depth-1]);
      std::vector< std::pair<int,int> > openPanels2(m_NumberOfBoxesPerLevel[m_Depth-1]);
      std::vector< std::pair<int,int> >* nextOpenPanels;
      std::vector< std::pair<int,int> >* currentOpenPanels;
      int currentOpenPanelsSize;
      int nextOpenPanelsSize;

      enum { EVALUATE_FAR_FIELD,
	     EVALUATE_DIRECT,
	     DEFER_TO_NEXT_LEVEL };

      unsigned int whatToDo;

      //for(int i=0;i<m_NumberOfPoints;++i)
      {

	 //const ScalarType* const  sourcepoint = &m_Points[i][0];
	 

//------------------------------------------------
	 // Evaluate extra polynomial term

	 result = this->EvaluateExtraPolynomialTerms(sourcepoint);
	 
	 //std::cout << result[i] << " ";


//------------------------------------------------


	 openPanels1[0] = std::make_pair(0,64);
	 currentOpenPanelsSize = 1;

	 currentOpenPanels = &openPanels1;
	 nextOpenPanels = &openPanels2;

	 // Evaluate Far Field
	 for(int k=2;
	     k<=m_Depth;
	     ++k,currentOpenPanelsSize=nextOpenPanelsSize,
		std::swap(currentOpenPanels,nextOpenPanels))
	 {

	    nextOpenPanelsSize = 0;

	    const std::vector<PointType>& centerArray = m_CenterPerBox[k];
	    const PolynomialArrayType& polyArray = m_FarFieldExpansions[k];
	    //const std::valarray<ScalarType>& M = m_M[k];
	    const std::valarray<ScalarType>& R2 = m_R2[k];
	    //const std::valarray<ScalarType>& E = m_Error2KMRK[k];
	    const std::valarray<int>& boxToPointMap = m_BoxToPointMap[k];


	    for(int jj=0;jj<currentOpenPanelsSize;++jj)
	    {

	       const int second = (*currentOpenPanels)[jj].second;

	       for(int j=(*currentOpenPanels)[jj].first; j<second; ++j)
	       {
	       

		  const PointType& centerOfBox = centerArray[j];
		     
		  // shift the point so it's relative to the center of the
		  // box
		  displacement[0] = sourcepoint[0] - centerOfBox[0];
		  displacement[1] = sourcepoint[1] - centerOfBox[1];
		  displacement[2] = sourcepoint[2] - centerOfBox[2];

	       
		  // Determine if we should evaluate the far field or the
		  // basis functions directly to compute the influence from
		  // this panel. The variable UseFarField will be set to
		  // true if we should use the far field, otherwise it will
		  // be set to false.  Also, if UseFarField is set to true,
		  // determine to what degree we should use for the
		  // evaluation


		  displacement[3] = (displacement[0]*displacement[0]+
				     displacement[1]*displacement[1]+
				     displacement[2]*displacement[2]);

		  //const ScalarType distance = sqrt(displacement[3]);
		  const ScalarType distance2 = displacement[3];


		  // Given a certain desired accuracy and the distance form
		  // the panel center, we now need to compute the degree of
		  // polynomial that will give us the desired accuracy.

		  // If we're too close then don't evaluate far field.
		  if(distance2 <= m_WellSeparatedFactor2*R2[j])
		  {
		     if(k == m_Depth)
		     {
			// If we're at the lowest level, just evaluate direct
			whatToDo = EVALUATE_DIRECT;
		     }
		     else
		     {
			whatToDo = DEFER_TO_NEXT_LEVEL;
		     }
		  }
		  else
		  {

		     // Divide the distance away from the point by the
		     // radius. Subtract this value from m_PplusK

		     //degree = m_PplusK - static_cast<int>(m_DecreaseFactor *
		     //				  sqrt(distance2/R2[j]));
		     if(R2[j] > 1e-8)
			degree =  	m_PplusK - static_cast<int>(
			   m_DecreaseFactor * sqrt(distance2/R2[j]) +
			   m_DecreaseFactor2);
		     else
			degree =  	m_PplusK - static_cast<int>(
			   /*m_DecreaseFactor * sqrt(distance2/R2[j]) +*/
			   m_DecreaseFactor2);

		     // clamp to min degree
		     if(degree < m_MinDegree)
			degree = m_MinDegree;
		     // clamp to max degree
		     else if(degree > m_PplusK)
			degree = m_PplusK;

//		     	  std::cout << degree << std::endl;

		     // Now that we know the degree we must choose among 3
		     // options. If the number of points in the box is less than
		     // the cutoff, simply evaluate directly. If not, then if the
		     // degree is less than m_PplusK then compute the far
		     // field. If the degree is equal to m_PlusK then (since the
		     // far field cannot provide us with a good approximation) if
		     // we are at the lowest level just evaluate
		     // direct. Otherwise defer to the next level.
		     const int numberOfPoints = 
			boxToPointMap[j+1] - boxToPointMap[j];

		     //if(0)
		     if(numberOfPoints < m_NumberCutoffPoints[degree])
		     {
			whatToDo = EVALUATE_DIRECT;//PRINTND(numberOfPoints<<" "<<degree);
		     }
		     else //if(degree <= m_PplusK+1)
		     {
			whatToDo = EVALUATE_FAR_FIELD;
		     }
		  }

		  switch(whatToDo)
		  {
		     case EVALUATE_FAR_FIELD:
		     {
			this->EvaluateFarFieldPolynomial(
			   displacement, polyArray[j], degree, result);
			break;
		     }
		     case EVALUATE_DIRECT:
		     {
			
			this->EvaluateSumBasicFunctionsNew(sourcepoint,
							   &boxToPointMap[j],
							   result);
			break;
		     }
		     case DEFER_TO_NEXT_LEVEL:
		     {
			const int tmp = ChildFirst(j);
			(*nextOpenPanels)[nextOpenPanelsSize++] = 
			   std::make_pair(tmp,tmp+8);
			break;
		     }
		     default:
		     {
			abort();
		     }
		  }

	       }
	    }
	 }
      } // end for m_NumberOfPoints


      return result;

   }//end function





/////////////////////////////////////////////////////////////////

   // Evaluate the sum directly by summing over all the terms.
   // Mostly used for testing.
   ScalarType
   evaluator::
   EvaluateDirect( const ScalarType thisPoint[3] ) const
   {
      const int numberOfPoints = m_NumberOfPoints;
      
      // Iterate over points
      ScalarType result = 0.0;
      for(int m=0;m<numberOfPoints;m++)
      {

	 result += m_Coefficients[m] * 
	    this->BasicFunction(thisPoint,&m_Points[m][0]);

      }

      // Add the polynomial part

      return result + this->EvaluateExtraPolynomialTerms(thisPoint);

   }


   // Evaluate the sum directly at every point by summing over all the terms.
   // Mostly used for testing.
   void
   evaluator::
   EvaluateAtAllSourcePointsDirect(ScalarType* const result) const
   {

      for(int i=0;i<m_NumberOfPoints;++i)
      {

	 result[i] = this->EvaluateDirect(&m_Points[i][0]);

      }
   }




////////////////////////////////////////////////////////////////


void evaluator::EvaluateGrid( ScalarType* result, 
			      const ScalarType spacing[3],
			      const ScalarType origin[3],
			      const int size[3]) const
{
   const ScalarType dx = spacing[0];
   const ScalarType dy = spacing[1];
   const ScalarType dz = spacing[2];
   const ScalarType xorig = origin[0];
   const ScalarType yorig = origin[1];
   const ScalarType zorig = origin[2];
   const int xsize = size[0];
   const int ysize = size[1];
   const int zsize = size[2];

   int n=0;
   ScalarType point[3];

   point[2] = zorig;
   for(int k=0;k<zsize;k++)
   {
      point[1] = yorig;
      for(int j=0;j<ysize;j++)
      {
	 point[0] = xorig;
	 for(int i=0;i<xsize;i++)
	 {
	    result[n++] = this->Evaluate(point);
	    point[0] += dx;
	 }
	 point[1] += dy;
      }
      point[2] += dz;
   }
}





}



#endif


#if 0

		     ScalarType pt[3] = {0,0,0};
		     ScalarType disp[4];
		     
		  for(int q=0;q<100;q++)
		  {
		     
		  disp[0] = pt[0] - centerOfBox[0];
		  disp[1] = pt[1] - centerOfBox[1];
		  disp[2] = pt[2] - centerOfBox[2];
		  disp[3] = (disp[0]*disp[0]+
			     disp[1]*disp[1]+
			     disp[2]*disp[2]);

		  ScalarType r = 0.0;
		  this->EvaluateSumBasicFunctionsNew(pt,
						     &boxToPointMap[j],
						     r);
		  ScalarType r2 = 0.0;
		  this->EvaluateFarFieldPolynomial(
		     disp, polyArray[j],11/* degree*/, r2);

		  pt[0]+=.01;
		  pt[1]+=.01;
		  pt[2]+=.01;

		     PRINT(fabs(r-r2));
		     PRINT(r);
		     PRINT(r2);
		     PRINT(sqrt(disp[3]));
		     DISPLAY("\n");
		  }
//		  if(0)
//		  if(fabs(r-r2) > .000000001 && whatToDo == 0 && j !=0 &&j!=9 )
		  {
//		     PRINT(fabs(r-r2));
//		     PRINT(r);
//		     PRINT(r2);
		     PRINT(distance);
		     PRINT(i);
		     PRINT(j);
		     DISPLAY(sourcepoint[0]);
		     DISPLAY(sourcepoint[1]);
		     DISPLAY(sourcepoint[2]);
		     DISPLAY("\n");
		     PRINT(degree);
		     DISPLAY(displacement[0]);
		     DISPLAY(displacement[1]);
		     DISPLAY(displacement[2]);
		     DISPLAY("\n");
		     PRINT(displacement[1]/displacement[2]);
		     DISPLAY(centerOfBox[0]);
		     DISPLAY(centerOfBox[1]);
		     DISPLAY(centerOfBox[2]);
		     DISPLAY("\n");
		     DISP("\n");


		  }
		  

		  abort();
#endif
