#ifndef __rbfmatrix_h
#define __rbfmatrix_h


#include <cmath>
#include <cassert>


#include "macros.h"
#include "evaluator.h"

namespace fmi
{


// This matrix represents the square rbf operator matrix
// To use this class one setup function must be called:
// SetupMatrix(...)

// This matrix is templated over the point type used
// for setting up the matrix

// This matrix doesn't actually store its components;
// instead when a method is called requesting data, 
// the data is generated on the fly
class rbfmatrix
{
   public:
   
      //typedef float ScalarType;

      // Set the matrix: 
      void SetupMatrix(const ScalarType* landmarks, const int numberOfPoints,
		       const int degree, const int k, const ScalarType tau, 
		       const ScalarType lambda = 0.0);


      template<class TVector>
      TVector mult(const TVector& v) const;

      void mult(const ScalarType* v,ScalarType* result) const;
      
      // Fast multiply with the fast multipole method
      template<class TVector>
      TVector mult_fast(const TVector& v) const;

	  
      ScalarType operator()(int i,int j) const;
   
    
      int size1() const { return num_rows; }
   
      int size2() const { return num_rows; }

      
   
      ScalarType BasicFunction(const ScalarType point[3], 
			       const ScalarType center[3]) const;


   private:


      // The list of source landmarks
      const ScalarType* m_SourceLandmarks;

      int num_rows; // since matrix is square there is no num_cols variable
      int m_NumberOfPoints;
      int m_PolynomialDegree;
      ScalarType m_Tau;
      ScalarType m_Tau2;
      int m_K;
      ScalarType m_Lambda;
//      evaluator eval;
	 
};


      // Set the matrix: 
   inline
   void 
   rbfmatrix::SetupMatrix(const ScalarType* landmarks, const int numberOfPoints,
			  const int degree, const int k, const ScalarType tau, 
			  const ScalarType lambda)
   {
      m_Tau = tau; 
      m_Tau2 = tau*tau; 
      m_K = k; 

      m_Lambda = lambda;

      if(lambda < 0.0)
      {
	 std::cout << "Error: lambda must be >= 0.0"
		   << std::endl;
	 abort();
      }

      if(lambda > 0.0 && (k != 1 || tau != 0.0))
      {
	 std::cout << "Error: lambda can only be nonzero "
	    "if k == 1 and tau == 0." << std::endl;
	 abort();
      }

      m_SourceLandmarks = landmarks; 
      m_NumberOfPoints = numberOfPoints;
      assert(m_NumberOfPoints > 0);

      assert(degree >= -1 && degree <= 2);
      m_PolynomialDegree = degree;

      int numberPolynomialTerms=0;
      switch(m_PolynomialDegree)
      {
	 case -1:
	    numberPolynomialTerms = 0;
	    break;
	 case 0:
	    numberPolynomialTerms = 1;
	    break;
	 case 1:
	    numberPolynomialTerms = 4;
	    break;
	 case 2:
	    numberPolynomialTerms = 10;
	    break;
      }

      num_rows = m_NumberOfPoints + numberPolynomialTerms;

   }


   template<class TVector>
   inline TVector rbfmatrix::mult(const TVector& v) const
   {
      TVector result(num_rows);

      this->mult(&v[0],&result[0]);

      return result;
   }


   template<class TVector>
   inline TVector rbfmatrix::mult_fast(const TVector& v) const
   {
/*
   {
      TVector result(num_rows);

      this->mult(&v[0],&result[0]);

      return result;
   }
*/

      TVector result(num_rows);

      //this->mult(&v[0],&result[0]);

      evaluator eval;

      //int pp=15;
      //int dep=2;

      eval.SetPointSet( m_SourceLandmarks, m_NumberOfPoints );
      eval.SetCoefficients( &v[0], m_NumberOfPoints );
      //ScalarType extent[6] = {0, 1, 0, 1, 0, 1};
      //eval.SetBoundingBox(extent);
      if(m_PolynomialDegree >= 0)
      {
	 //abort();
	 eval.SetExtraPolynomial(m_PolynomialDegree,&v[m_NumberOfPoints]);
      }
      else
	 eval.SetExtraPolynomial(-1,0);
      eval.SetP(15);
      eval.SetK(m_K);
      //eval.SetTau( std::pow(1.0/m_NumberOfPoints,1.0/3.0) );
      eval.SetTau( m_Tau );
      //eval.SetDepth(dep);
      eval.SetDepthToOptimal();
      eval.SetWellSeparatedFactor(2.0);
      eval.SetDecreaseFactor(1);
      eval.SetDecreaseFactor2(0);
      eval.SetMinDegree(6);


      eval.Update();

      eval.EvaluateAtAllSourcePoints(&result[0]);

      if(0) // for debugging only
      {
	 double average_error = 0;
	 double max_error = 0.0;
	 double min_error = 1e30;
	 for(int i=0;i<m_NumberOfPoints;i++)
	 {

	    double r = result[i];
	    double r2 = eval.EvaluateDirect(&m_SourceLandmarks[3*i]);
	
	    average_error += fabs(r-r2);

	    if( fabs(r-r2) > max_error )
	       max_error = fabs(r-r2);
	    if( fabs(r-r2) < min_error )
	       min_error = fabs(r-r2);
      
	 }

	 average_error /= m_NumberOfPoints;
	
	 std::cout << std::endl << "average_error: " << average_error << std::endl;
	 std::cout << "max_error: " << max_error << std::endl;
	 std::cout << "min_error: " << min_error << std::endl;
      }


      for(int i=0;i<m_NumberOfPoints;i++)
      {
	 result[i] += m_Lambda * v[i] * static_cast<double>(m_NumberOfPoints);
      }


      return result;
   }




   inline void rbfmatrix::mult(const ScalarType* v,ScalarType* result) const
   {

      int i=0;
      for(;i<m_NumberOfPoints;i++)
      {
	       
	 const ScalarType* thisPoint = &m_SourceLandmarks[3*i];
	  
	 // Iterate over points

	 //ScalarType temp = 0.0;
	 ScalarType temp = m_Lambda * v[i] * static_cast<double>(m_NumberOfPoints);

	 int r = 0;
	 for( ;r<m_NumberOfPoints;++r)
	 {
	    temp += 
	       v[r]*this->BasicFunction(thisPoint,&m_SourceLandmarks[3*r]);
	 }
	 
      
	 // Add the polynomial part
	 switch(m_PolynomialDegree)
	 {
	    case 2:
	       // The monomial basis is:
	       // 1, x, x^2, y, yx, y^2, z, zx, zy, z^2
	       // 
	       temp += v[r];
	       temp += v[++r]*thisPoint[0];
	       temp += v[++r]*thisPoint[0]*thisPoint[0];
	       temp += v[++r]*thisPoint[1];
	       temp += v[++r]*thisPoint[1]*thisPoint[0];
	       temp += v[++r]*thisPoint[1]*thisPoint[1];
	       temp += v[++r]*thisPoint[2];
	       temp += v[++r]*thisPoint[2]*thisPoint[0];
	       temp += v[++r]*thisPoint[2]*thisPoint[1];
	       temp += v[++r]*thisPoint[2]*thisPoint[2];
	    
	       break;
	    
	    case 1:
	       // The monomial basis is:
	       // 1, x, y, z
	       // 
	       temp += v[r];
	       temp += v[++r]*thisPoint[0];
	       temp += v[++r]*thisPoint[1];
	       temp += v[++r]*thisPoint[2];
	    
	       break;
	    
	    case 0:
	       // The monomial basis is:
	       // 1
	       // 
	       temp += v[r];
	    
	       break;
	    
	    case -1:
	       // do nothing
	       break;
	    
	 }
	 
	 result[i] = temp;
      }


      // add the bottom few rows of the matrix (corresponding
      // to the polynomial terms)
      for(;i<num_rows;i++)
      {

	 ScalarType temp = 0.0;

	 for(int r = 0;r<m_NumberOfPoints;++r)
	 {
	    temp += v[r] * (*this)(i,r);
	 }

	 result[i] = temp;

      }
   }
      
	  
   inline ScalarType rbfmatrix::operator()(int i,int j) const
   {

      assert(i >= 0);
      assert(i < num_rows);
      assert(j >= 0);
      assert(j < num_rows);
      

      
      if(i<m_NumberOfPoints && j<m_NumberOfPoints)
      {
	 if( i==j )
	    return m_Lambda + m_Tau;

	 return this->BasicFunction(&m_SourceLandmarks[3*i], 
				    &m_SourceLandmarks[3*j]);
      }
      else if(i>=m_NumberOfPoints && j>=m_NumberOfPoints)
      {
	 return 0.0;
      }
      else 
      {
	 if(j<m_NumberOfPoints)
	 {
	    //swap i and j
	    int tmp = i;
	    i = j;
	    j = tmp;
	 }
	 
	      
	 const ScalarType* sourcepoint  = &m_SourceLandmarks[3*i];

	 // Set the polynomial part
	 switch(m_PolynomialDegree)
	 {
	    case 2:
	       // The monomial basis is:
	       // 1, x, x^2, y, yx, y^2, z, zx, zy, z^2

	       j -= m_NumberOfPoints;
		  
	       switch(j)
	       {
		  case 0:
		     return 1.0;
		  case 1:
		     return sourcepoint[0];
		  case 2:
		     return sourcepoint[0]*sourcepoint[0];
		  case 3:
		     return sourcepoint[1];
		  case 4:
		     return sourcepoint[1]*sourcepoint[0];
		  case 5:
		     return sourcepoint[1]*sourcepoint[1];
		  case 6:
		     return sourcepoint[2];
		  case 7:
		     return sourcepoint[2]*sourcepoint[0];
		  case 8:
		     return sourcepoint[2]*sourcepoint[1];
		  case 9:
		     return sourcepoint[2]*sourcepoint[2];
	       }

	    case 1:
	       // The monomial basis is:
	       // 1, x, y, z
	       // 
	       j -= m_NumberOfPoints;

	       switch(j)
	       {

		  case 0:
		     return 1.0;
		  case 1:
		     return sourcepoint[0];
		  case 2:
		     return sourcepoint[1];
		  case 3:
		     return sourcepoint[2];
	       }	
  
	    case 0:
	       // The monomial basis is:
	       // 1
	       // 
	       return 1.0;

	 }
      }

	 
      // if we reach here something's wrong so abort
      abort();
      return 0.0;
	 
   }
   
    
   inline ScalarType rbfmatrix::BasicFunction(const ScalarType point[3], 
					      const ScalarType center[3]) const
   {
      const ScalarType x = point[0] - center[0];
      const ScalarType y = point[1] - center[1];
      const ScalarType z = point[2] - center[2];
      return sqrt( std::pow(x*x + y*y + z*z + m_Tau2, m_K ) );

      //ScalarType val = x*x + y*y + z*z + m_Tau2;
      //return m_BasicFunctionTable[(int)(val/delta)];
	 
   }

   
}

#endif

